unit mxBNMathConv;

// Binuar 1.2.1
// Big integer number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{$I mxBNMath.inc}

interface

uses
  mxBNMathTypes;

procedure BN_InitString (var S: string; L: UIntX);
function  BN_ParseString (const S: string; var ASign: ShortInt): string;

procedure BN_Base2 (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_Base2 (var V: T_BNX; const S: string); overload;
function  BN_Base2 (const A: T_BNX): string; overload;

procedure BN_Base2x (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_Base2x (var V: T_BNX; const S: string); overload;
function  BN_Base2x (const A: T_BNX): string; overload;

procedure BN_Base8 (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_Base8 (var V: T_BNX; const S: string); overload;
function  BN_Base8 (const A: T_BNX): string; overload;

procedure BN_Base10 (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_Base10 (var V: T_BNX; const S: string); overload;
function  BN_Base10 (A: T_BNX): string; overload;

procedure BN_Base16 (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_Base16 (var V: T_BNX; const S: string); overload;
function  BN_Base16 (const A: T_BNX): string; overload;

procedure BN_Base256 (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_AnsiBase256 (var V: T_BNX; const S: AnsiString; ASign: ShortInt); overload;
function  BN_Base256 (const A: T_BNX): string; overload;
function  BN_AnsiBase256 (const A: T_BNX): AnsiString; overload;

procedure BN_String (var V: T_BNX; const S: string; ASign: ShortInt); overload;
procedure BN_AnsiString (var V: T_BNX; const S: AnsiString; ASign: ShortInt); overload;
function  BN_String (const A: T_BNX): string; overload;
function  BN_AnsiString (const A: T_BNX): AnsiString; overload;

implementation

uses
  SysUtils, mxBNMath, mxBNMathUtils, mxBNMathProcs;

{
  '0'..'9' -> 0..9
}
function CharToInt (C: Char): UIntX;
{$IFDEF CPUX64}
// rcx | C
asm
       .NOFRAME

        {$IFDEF UNICODE}
        and     rcx, $FFFF
        {$ELSE}
        and     rcx, $FF
        {$ENDIF}
        lea     rax, [rcx - $30]
end;
{$ELSE}
// eax | C
asm
        {$IFDEF UNICODE}
        and     eax, $FFFF
        {$ELSE}
        and     eax, $FF
        {$ENDIF}
        lea     eax, [eax - $30]
end;
{$ENDIF}

{
}
procedure BN_InitString (var S: string; L: UIntX);
var
  I : LoopUInt;
begin
  SetLength (S, L);
  for I := 1 to L do S[I] := '0';
end;

{
}
function BN_ParseString (const S: string; var ASign: ShortInt): string;
var
  L  : IntX;
  C  : Char;
  PS : string;
begin
  ASign := 1;

  PS := Trim (S);
  L := Length (S);

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  C := PS[1];

  if C = '-' then begin
    ASign := -1;
    Delete (PS, 1, 1);
    Result := Trim (PS);
    Exit;
  end;

  if C = '+' then begin
    Delete (PS, 1, 1);
    Result := Trim (PS);
    Exit;
  end;

  Result := PS;
end;

{
       
     
      

  Get a number from the binary notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  count -   (number of characters)
  s     -   (address of a string)

  110011.101 -> 101
}
function Base2_UIntX_asm (count: UIntX; var s: Pointer): UIntX;
{$IFDEF CPUX64}
// rcx | count  |     |
// rdx | @s     | r8  | s
//     | bit    | r9  |
//     | result | rax |
asm
       .NOFRAME

        mov     r8, [rdx]

        xor     rax, rax
        xor     r9, r9
        inc     r9 // r9 = 1
  @loop:
        cmp     byte ptr [r8], '0' // ? char = '0'
        jz      @0
        cmp     byte ptr [r8], '1' // ? char = '1'
        jz      @1
        xor     r8, r8 // invalid char
        jmp     @r
  @1:   or      rax, r9
  @0:   {$IFDEF UNICODE}
        sub     r8, 2
        {$ELSE}
        dec     r8
        {$ENDIF}
        shl     r9, 1
  @next:
        dec     rcx
        jnz     @loop

  @r:
        mov     [rdx], r8
end;
{$ELSE}
// eax | count  | ecx |
// edx | @s     | esi | s
//     | bit    | ebx |
//     | result | eax |
asm
        push    ebx
        push    esi

        mov     ecx, eax
        mov     esi, [edx]

        xor     eax, eax
        xor     ebx, ebx
        inc     ebx // ebx = 1
  @loop:
        cmp     byte ptr [esi], '0' // ? char = '0'
        jz      @0
        cmp     byte ptr [esi], '1' // ? char = '1'
        jz      @1
        xor     esi, esi // invalid char
        jmp     @r
  @1:   or      eax, ebx
  @0:   {$IFDEF UNICODE}
        sub     esi, 2
        {$ELSE}
        dec     esi
        {$ENDIF}
        shl     ebx, 1
  @next:
        dec     ecx
        jnz     @loop

  @r:
        mov     [edx], esi

        pop     esi
        pop     ebx
end;
{$ENDIF}

{
       
     
      

  Get a number from the binary notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  110011.101 -> 101
}
procedure BN_Base2 (var V: T_BNX; const S: string; ASign: ShortInt);
var
  I      : LoopInt;
  L      : IntX;
  P      : Pointer;
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CXBits);
  BN_Init (V, ASign, VCount);

  P := PChar(S) + L - 1;
  for I := 0 to V.HIndex do begin
    V.Data[I] := Base2_UIntX_asm (CXBits, P);
    if P = nil then Break;
  end;
  BN_Pack (V);
end;

{
       
     
      

  Get a number from the binary notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  110011.101 -> 101
}
procedure BN_Base2 (var V: T_BNX; const S: string);
var
  PS    : string;
  VSign : ShortInt;
begin
  PS := BN_ParseString (S, VSign);
  BN_Base2 (V, PS, VSign);
end;

{
       
  Convert the number to binary notation

  count -   (number of characters)
  i     -  (number)
  s     -   (address of a string)
}
procedure UIntX_Base2_asm (count: UIntX; i: UIntX; var s: Pointer);
{$IFDEF CPUX64}
// rcx | count |     |
// rdx | i     |     |
// r8  | @s    | r10 | s
asm
       .NOFRAME

        mov     r10, [r8]
  @loop:
        shl     rdx, 1
        jnc     @0
        {$IFDEF UNICODE}
  @1:   mov     word ptr [r10], '1'
        jmp     @inc
  @0:   mov     word ptr [r10], '0'
  @inc: add     r10, 2
        {$ELSE}
  @1:   mov     byte ptr [r10], '1'
        jmp     @inc
  @0:   mov     byte ptr [r10], '0'
  @inc: inc     r10
        {$ENDIF}
  @next:
        dec     rcx
        jnz     @loop

        mov     [r8], r10
end;
{$ELSE}
// eax | count |     |
// edx | i     |     |
// ecx | @s    | edi | s
asm
        push    edi

        mov     edi, [ecx]
  @loop:
        shl     edx, 1
        jnc     @0
        {$IFDEF UNICODE}
  @1:   mov     word ptr [edi], '1'
        jmp     @inc
  @0:   mov     word ptr [edi], '0'
  @inc: add     edi, 2
        {$ELSE}
  @1:   mov     byte ptr [edi], '1'
        jmp     @inc
  @0:   mov     byte ptr [edi], '0'
  @inc: inc     edi
        {$ENDIF}
  @next:
        dec     eax
        jnz     @loop

        mov     [ecx], edi

        pop     edi
end;
{$ENDIF}

{
       
  Convert the number to binary notation
}
function BN_Base2 (const A: T_BNX): string;
var
  I      : LoopInt;
  Imax   : IntX;
  L      : IntX;
  P      : Pointer;
  VCount : IntX;
  VBits  : UIntX;
  VmBits : UIntX;
  VData  : UIntX;
begin
  VBits := BN_DataBits (A, VCount);
  if VBits = 0 then begin
    Result := '';
    Exit;
  end;

  L := VBits;
  VmBits := VBits mod CXBits;

  SetLength (Result, L);
  P := PChar(Result);

  Imax := VCount - 1;
  if VmBits = 0 then begin
    for I := Imax downto 0 do begin
      UIntX_Base2_asm (CXBits, A.Data[I], P);
    end;
  end else begin
    VData := A.Data[Imax] shl (CXBits - VmBits);
    UIntX_Base2_asm (VmBits, VData, P);
    for I := Imax - 1 downto 0 do begin
      UIntX_Base2_asm (CXBits, A.Data[I], P);
    end;
  end;
end;

{
       
    

  Convert the number to binary notation
  Draw leading zeros

  101 -> '00000000000000000000000000000101'
}
function BN_Base2_LZ (const A: T_BNX): string;
var
  I      : LoopInt;
  L      : IntX;
  P      : Pointer;
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 6; // VCount * 64;
  {$ELSE}
  L := VCount shl 5; // VCount * 32;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  SetLength (Result, L);
  P := PChar(Result);

  for I := VCount - 1 downto 0 do begin
    UIntX_Base2_asm (CXBits, A.Data[I], P);
  end;
end;

{
        
     
      

  Get a number from the extended binary notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  1111.1-00 -> 1-00
}
procedure BN_Base2x (var V: T_BNX; const S: string; ASign: ShortInt);

  function Char2xToInt_(C: Char; var AError: Boolean): IntX;
  begin
    Result := 0;
    AError := true;

    if C = '0' then begin
      AError := false;
    end else
    if C = '1' then begin
      Result := +1;
      AError := false;
    end else
    if C = '-' then begin
      Result := -1;
      AError := false;
    end;
  end;

var
  I      : LoopInt;
  L      : IntX;
  B      : IntX;
  P      : IntX;
  V2n    : T_BNX;
  VError : Boolean;
begin
  BN_AssignZero (V);

  L := Length (S);
  if L <= 0 then Exit;

  B := -1;
  for I := L downto 1 do begin
    Inc (B);
    P := Char2xToInt_(S[I], VError);
    if VError then Break; // invalid char
    if P = 0 then Continue;

    BN_AssignBit (V2n, B);
    if P > 0 then begin
      BN_add_(V, V2n);
    end else begin
      BN_sub_(V, V2n);
    end;
  end;

  V.Sign := ASign;

  BN_LFree (V2n);
end;

{
        
  Get a number from the extended binary notation
}
procedure BN_Base2x (var V: T_BNX; const S: string);
var
  PS    : string;
  VSign : ShortInt;
begin
  PS := BN_ParseString (S, VSign);
  BN_Base2x (V, PS, VSign);
end;

{
        
  Convert the number to extended binary notation

  100    ->           10-00100 = 2^7 - 2^5 + 2^2
  1000   ->        100000--000 = 2^10 - 2^4 - 2^3
  10000  ->     10100-00010000 = 2^13 + 2^11 - 2^8 + 2^4
  100000 -> 10-000100-0--00000 = 2^17 - 2^15 + 2^11 - 2^8 - 2^6 - 2^5
}
function BN_Base2x (const A: T_BNX): string;
var
  L     : UIntX;
  VSign : Integer;
  VBits : UIntX;
  VChar : Char;
  Va    : T_BNX;
  // V1    : T_BNX;
  VS1   : T_BNX;
  V2    : T_BNX;
  VS2   : T_BNX;
begin
  VBits := BN_DataBits (A);
  if VBits = 0 then begin
    Result := '';
    Exit;
  end;

  (*
  BN_AssignBit (V1, VBits - 1);
  BN_sub (VS1, A, V1);
  // *)
  BN_ClearBit (VS1, A, VBits - 1);
  // ----
  BN_AssignBit (V2, VBits);
  BN_sub (VS2, V2, A);
  // ----
  if BN_DataCompare (VS1, VS2) <= 0 then begin
    L := VBits;
    VSign := +1;
    BN_copy (Va, VS1);
  end else begin
    L := VBits + 1;
    VSign := -1;
    BN_copy (Va, VS2);
  end;
  // ----
  BN_InitString (Result, L);
  Result[1] := '1';

  VBits := BN_DataBits (Va);
  while VBits <> 0 do begin
    (*
    BN_AssignBit (V1, VBits - 1);
    BN_sub (VS1, Va, V1);
    // *)
    BN_ClearBit (VS1, Va, VBits - 1);
    // ----
    BN_AssignBit (V2, VBits);
    BN_sub (VS2, V2, Va);
    // ----
    if VSign > 0 then VChar := '1' else VChar := '-';
    if BN_DataCompare (VS1, VS2) <= 0 then begin
      Result[L - VBits + 1] := VChar;
      BN_copy (Va, VS1);
    end else begin
      Result[L - VBits] := VChar;
      VSign := -VSign;
      BN_copy (Va, VS2);
    end;
    // if VBits = 1 then Break;
    VBits := BN_DataBits (Va);
  end;

  BN_LFree (Va);
  // BN_LFree (V1);
  BN_LFree (VS1);
  BN_LFree (V2);
  BN_LFree (VS2);
end;

{
       
     
      

  Get a number from the octal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  7654.3210 -> 3210
}
procedure BN_Base8 (var V: T_BNX; const S: string; ASign: ShortInt);
var
  I       : LoopInt;
  L       : IntX;
  P       : UIntX;
  C       : UIntX;
  VBits   : IntX;
  VCount  : IntX;

  VIndex  : UIntX;
  VmIndex : UIntX;
  VPIndex : UIntX; // previous value
  VCIndex : UIntX; // current value
  VShiva  : UIntX; // shift value
  VPValue : UIntX; // previous value
  VCValue : UIntX; // current value
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VBits := L * 3;
  VCount := cdiv (VBits, CXBits);
  BN_Init (V, ASign, VCount);

  C := 3;
  VPIndex := 0;
  VCIndex := 0;
  VmIndex := C mod CXBits;

  VPValue := 0;
  VCValue := 0;

  for I := L downto 1 do begin
    P := CharToInt (S[I]);
    if P > 7 then Break; // invalid char

    if VmIndex < 3 then begin
      VShiva := 3 - VmIndex;
      VCValue := VCValue or (P shr VShiva);
      VPValue := VPValue or (P shl (CXBits - VShiva));
      V.Data[VCIndex] := VCValue;
      V.Data[VPIndex] := VPValue;
    end else begin
      VShiva := VmIndex - 3;
      VCValue := VCValue or (P shl VShiva);
      V.Data[VCIndex] := VCValue;
    end;
    if I = 1 then Break;

    Inc (C, 3);

    VIndex := div_mod (C, CXBits, VmIndex);
    if VIndex <> VCIndex then begin
      VPIndex := VCIndex;
      VPValue := VCValue;
      VCIndex := VIndex;
      VCValue := 0;
    end;
  end;
  BN_Pack (V);
end;

{
       
     
      

  Get a number from the octal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  7654.3210 -> 3210
}
procedure BN_Base8 (var V: T_BNX; const S: string);
var
  PS    : string;
  VSign : ShortInt;
begin
  PS := BN_ParseString (S, VSign);
  BN_Base8 (V, PS, VSign);
end;

{
       
  Convert the number to octal notation
}
function BN_Base8 (const A: T_BNX): string;
var
  I       : LoopUInt;
  L       : UIntX;
  C       : UIntX;
  V       : UIntX;
  VBits   : UIntX;

  VIndex  : UIntX;
  VmIndex : UIntX;
  VCIndex : UIntX; // current value
  VShiva  : UIntX; // shift value
  VPValue : UIntX; // previous value
  VCValue : UIntX; // current value
begin
  VBits := BN_DataBits (A);
  if VBits = 0 then begin
    Result := '';
    Exit;
  end;

  L := cdiv (VBits, 3);
  SetLength (Result, L);

  C := 3;
  VCIndex := 0;
  VmIndex := C mod CXBits;

  VPValue := 0;
  VCValue := A.Data[0];

  for I := L downto 1 do begin
    if VmIndex < 3 then begin
      VShiva := 3 - VmIndex;
      V := (VCValue shl VShiva) or (VPValue shr (CXBits - VShiva));
    end else begin
      VShiva := VmIndex - 3;
      V := (VCValue shr VShiva);
    end;
    Result[I] := CHR($30 + (V and 7));
    if I = 1 then Break;

    Inc (C, 3);

    VIndex := div_mod (C, CXBits, VmIndex);
    if VIndex <> VCIndex then begin
      VPValue := VCValue;
      VCIndex := VIndex;
      VCValue := A.Data[VCIndex];
    end;
  end;
end;

{
       
     
      

  Get a number from the decimal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  12345.6789 -> 6789
}
procedure BN_Base10 (var V: T_BNX; const S: string; ASign: ShortInt);
var
  I    : LoopInt;
  L    : IntX;
  P    : UIntX;
  Vn   : T_BNX;
  V10n : T_BNX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  P := CharToInt (S[L]);
  if P > 9 then begin // invalid char
    BN_AssignZero (V);
    Exit;
  end;

  BN_AssignInt (V, 1, P);

  BN_AssignInt (V10n, 1, 10);
  for I := L - 1 downto 1 do begin
    P := CharToInt (S[I]);
    if P > 9 then Break; // invalid char

    BN_mul_int (Vn, V10n, 1, P);
    BN_add_(V, Vn);

    if I = 1 then Break;
    BN_copy (Vn, V10n);
    BN_mul_int (V10n, Vn, 1, 10); // V10n = V10n * 10
  end;

  V.Sign := ASign;

  BN_LFree (Vn);
  BN_LFree (V10n);
end;

{
       
  Get a number from the decimal notation
}
procedure BN_Base10 (var V: T_BNX; const S: string);
var
  PS    : string;
  VSign : ShortInt;
begin
  PS := BN_ParseString (S, VSign);
  BN_Base10 (V, PS, VSign);
end;

{
  10^n = 2^(Bits+1)
  n = (Bits+1) * lg(2)
  lg(2) = 0.30102999566398119521373889472449
        =   30103 / 100000
        =   301029996 / 1000000000
        =   3010299956639811953 / 10000000000000000000

  Result = ((Bits + 1) * 30103) div 100000
         = (Bits * 30103 + 30103) div 100000
}
function Get10n_asm (ABits: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | ABits | rax
asm
       .NOFRAME

     // xor     rdx, rdx
        mov     rax, rcx
     // mov     rcx, 30103
        mov     rcx, 3010299956639811953
        mul     rcx
        add     rax, rcx
        adc     rdx, 0
     // mov     rcx, 100000
        mov     rcx, 10000000000000000000
        div     rcx
end;
{$ELSE}
// eax | ABits
asm
     // xor     edx, edx
     // mov     ecx, 30103
        mov     ecx, 301029996
        mul     ecx
        add     eax, ecx
        adc     edx, 0
     // mov     ecx, 100000
        mov     ecx, 1000000000
        div     ecx
end;
{$ENDIF}

{
       
  Convert the number to decimal notation
}
function BN_Base10 (A: T_BNX): string;
var
  V10nn : T_BNX;

  procedure Get10n_(var A10n: T_BNX; var n: UIntX; const ABase: T_BNX);
  var
    VBits : UIntX;
  begin
    n := 0;

    VBits := BN_DataBits (ABase);
    if VBits <= 0 then Exit;

    n := Get10n_asm (VBits);
    BN_exp10 (A10n, 1, n); // A10n = 10^n

    while BN_DataCompare (ABase, A10n) < 0 do begin
      Dec (n);
      if n <= 0 then Break;
      BN_copy (V10nn, A10n);
      BN_div_int (A10n, V10nn, 1, 10);
    end;
  end;

  function AsString_(const V: T_BNX): string;
  begin
    Result := '0';
    if V.Count <> 1 then Exit;
    Result := IntToStr (V.Data[0]);
  end;

  function AsChar_(const V: T_BNX): Char;
  var
    V0 : UIntX;
  begin
    Result := '*';

    if V.Count <> 1 then begin
      // Result := '$';
      Exit;
    end;

    V0 := V.Data[0];
    if V0 > 9 then begin
      // Result := '#';
      Exit;
    end;

    Result := CHR($30 + V0);
  end;

var
  VCount : UIntX;
  n      : UIntX;
  V10n   : T_BNX;
  V_div  : T_BNX;
  V_mod  : T_BNX;
  V_a    : T_BNX;
begin
  Get10n_(V10n, n, A);
  if n <= 0 then begin
    Result := AsString_(A);
    BN_LFree (V10n);
    Exit;
  end;

  VCount := n + 1;
  BN_InitString (Result, VCount);

  BN_div_mod (V_div, V_mod, A, V10n);
  Result[VCount - n] := AsChar_(V_div);

  while true do begin
    Get10n_(V10n, n, V_mod);
    if n <= 0 then Break;

    BN_copy (V_a, V_mod);
    BN_div_mod (V_div, V_mod, V_a, V10n);
    Result[VCount - n] := AsChar_(V_div);
  end;

  if V_mod.Count = 1 then begin
    Result[VCount] := AsChar_(V_mod);
  end;

  if A.Sign < 0 then Result := '-' + Result;

  BN_LFree (V10n);
  BN_LFree (V10nn);
  BN_LFree (V_div);
  BN_LFree (V_mod);
  BN_LFree (V_a);
end;

{
       
     
      

  Get a number from the hexadecimal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  count -   (number of characters)
  s     -   (address of the string)

  123456789.ABCDEF -> ABCDEF
}
function Base16_UIntX_asm (count: UIntX; var s: Pointer): UIntX;
{$IFDEF CPUX64}
// rcx | count  | r8  |
// rdx | @s     | r9  | s
//     | char   | r10 |
//     | result | rax |
asm
       .NOFRAME

        mov     r8, rcx
        mov     r9, [rdx]

        xor     rax, rax
        xor     rcx, rcx
  @loop:
        xor     r10, r10
        mov     r10b, byte ptr [r9]
  @dg:  sub     r10b, '0'     // ? char in '0'..'9'
        cmp     r10b, 10
        jb      @dec
  @aUC: sub     r10b, 'A'-'0' // ? char in 'A'..'F'
        cmp     r10b, 6
        jb      @a
  @aLC: sub     r10b, 'a'-'A' // ? char in 'a'..'f'
        cmp     r10b, 6
        jb      @a
        xor     r9, r9 // invalid char
        jmp     @r
  @a:   add     r10b, 10
  @dec: shl     r10, cl
        or      rax, r10
        {$IFDEF UNICODE}
        sub     r9, 2
        {$ELSE}
        dec     r9
        {$ENDIF}
  @next:
        add     rcx, 4
        dec     r8
        jnz     @loop

  @r:
        mov     [rdx], r9
end;
{$ELSE}
// eax | count  | edi |
// edx | @s     | esi | s
//     | char   | ebx |
//     | result | eax |
asm
        push    ebx
        push    esi
        push    edi

        mov     edi, eax
        mov     esi, [edx]

        xor     eax, eax
        xor     ecx, ecx
  @loop:
        xor     ebx, ebx
        mov     bl, byte ptr [esi]
  @dg:  sub     bl, '0'     // ? char in '0'..'9'
        cmp     bl, 10
        jb      @dec
  @aUC: sub     bl, 'A'-'0' // ? char in 'A'..'F'
        cmp     bl, 6
        jb      @a
  @aLC: sub     bl, 'a'-'A' // ? char in 'a'..'f'
        cmp     bl, 6
        jb      @a
        xor     esi, esi // invalid char
        jmp     @r
  @a:   add     bl, 10
  @dec: shl     ebx, cl
        or      eax, ebx
        {$IFDEF UNICODE}
        sub     esi, 2
        {$ELSE}
        dec     esi
        {$ENDIF}
  @next:
        add     ecx, 4
        dec     edi
        jnz     @loop

  @r:
        mov     [edx], esi

        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
       
     
      

  Get a number from the hexadecimal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  123456789.ABCDEF -> ABCDEF
}
procedure BN_Base16 (var V: T_BNX; const S: string; ASign: ShortInt);
const
  {$IFDEF CPUX64}
  CNibbles = 16;
  {$ELSE}
  CNibbles = 8;
  {$ENDIF}
var
  I      : LoopInt;
  L      : IntX;
  P      : Pointer;
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CNibbles);
  BN_Init (V, ASign, VCount);

  P := PChar(S) + L - 1;
  for I := 0 to V.HIndex do begin
    V.Data[I] := Base16_UIntX_asm (CNibbles, P);
    if P = nil then Break;
  end;
  BN_Pack (V);
end;

{
       
     
      

  Get a number from the hexadecimal notation
  Scan the string from right to left
  Interrupt scanning when an invalid symbol

  123456789.ABCDEF -> ABCDEF
}
procedure BN_Base16 (var V: T_BNX; const S: string);
var
  PS    : string;
  VSign : ShortInt;
begin
  PS := BN_ParseString (S, VSign);
  BN_Base16 (V, PS, VSign);
end;

{
       
  Convert the number to hexadecimal notation

  count -   (number of characters)
  i     -  (number)
  s     -   (address of the string)
}
procedure UIntX_Base16_asm (count: UIntX; i: UIntX; var s: Pointer);
{$IFDEF CPUX64}
// rcx | count       |
// rdx | i           |
// r8  | @s          | r10 | s
//     | nibble char | rax |
asm
       .NOFRAME

        mov     r10, [r8]
        mov     r11, $07 // upper case
     // mov     r11, $27 // lower case
        xor     rax, rax
  @loop:
        mov     al, dl
        and     al, $0F
        cmp     al, 10
        jc      @a
        lea     rax, [rax + r11] // al >= 10
    @a:
        lea     rax, [rax + $30]

        {$IFDEF UNICODE}
        mov     word ptr [r10], ax
        sub     r10, 2
        {$ELSE}
        mov     byte ptr [r10], al
        dec     r10
        {$ENDIF}
        shr     rdx, 4
  @next:
        dec     rcx
        jnz     @loop

        mov     [r8], r10
end;
{$ELSE}
// eax | count       | ebx
// edx | i           |
// ecx | @s          | edi | s
//     | nibble char | eax |
asm
        push    ebx
        push    esi
        push    edi

        mov     ebx, eax
        mov     edi, [ecx]
        mov     esi, $07 // upper case
     // mov     esi, $27 // lower case
        xor     eax, eax
  @loop:
     // mov     al, dl
     // and     al, $0F
     // cmp     al, 10
     // sbb     al, $69
     // das

        mov     al, dl
        and     al, $0F
        cmp     al, 10
        jc      @a
        lea     eax, [eax + esi] // al >= 10
    @a:
        lea     eax, [eax + $30]

        {$IFDEF UNICODE}
        mov     word ptr [edi], ax
        sub     edi, 2
        {$ELSE}
        mov     byte ptr [edi], al
        dec     edi
        {$ENDIF}
        shr     edx, 4
  @next:
        dec     ebx
        jnz     @loop

        mov     [ecx], edi

        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
       
  Convert the number to hexadecimal notation
}
function BN_Base16 (const A: T_BNX): string;
const
  {$IFDEF CPUX64}
  CNibbles = 16;
  {$ELSE}
  CNibbles = 8;
  {$ENDIF}
var
  I         : LoopInt;
  Imax      : IntX;
  L         : IntX;
  P         : Pointer;
  VCount    : IntX;
  VNibbles  : UIntX;
  VmNibbles : UIntX;
begin
  VNibbles := BN_DataNibbles (A, VCount);
  if VNibbles = 0 then begin
    Result := '';
    Exit;
  end;

  L := VNibbles;
  VmNibbles := VNibbles mod CNibbles;

  SetLength (Result, L);
  P := PChar(Result) + L - 1;

  Imax := VCount - 1;
  if VmNibbles = 0 then begin
    for I := 0 to Imax do begin
      UIntX_Base16_asm (CNibbles, A.Data[I], P);
    end;
  end else begin
    for I := 0 to Imax - 1 do begin
      UIntX_Base16_asm (CNibbles, A.Data[I], P);
    end;
    UIntX_Base16_asm (VmNibbles, A.Data[Imax], P);
  end;
end;

{
       
    

  Convert the number to hexadecimal notation
  Draw leading zeros

  ABCD -> '0000ABCD'
}
function BN_Base16_LZ (const A: T_BNX): string;
const
  {$IFDEF CPUX64}
  CNibbles = 16;
  {$ELSE}
  CNibbles = 8;
  {$ENDIF}
var
  I      : LoopInt;
  L      : IntX;
  P      : Pointer;
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 4; // VCount * 16;
  {$ELSE}
  L := VCount shl 3; // VCount * 8;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  SetLength (Result, L);
  P := PChar(Result) + L - 1;

  for I := 0 to VCount - 1 do begin
    UIntX_Base16_asm (CNibbles, A.Data[I], P);
  end;
end;

{
     256-  
  Get a number from the 256-base notation
}
procedure BN_Base256 (var V: T_BNX; const S: string; ASign: ShortInt);
var
  L      : IntX;
  I      : LoopInt;
  P      : IntX;
  C      : Char;
  VC     : UIntX;
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CXBytes);
  BN_Init (V, ASign, VCount);

  P := L;
  for I := 0 to V.HIndex do begin
    VC := 0;

    {$IFDEF CPUX64}
    if P > 7 then begin C := S[P - 7]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 6 then begin C := S[P - 6]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 5 then begin C := S[P - 5]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 4 then begin C := S[P - 4]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 3 then begin C := S[P - 3]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 2 then begin C := S[P - 2]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 1 then begin C := S[P - 1]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 0 then begin C := S[P - 0]; VC := VC or ORD(C); end;
    {$ELSE}
    if P > 3 then begin C := S[P - 3]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 2 then begin C := S[P - 2]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 1 then begin C := S[P - 1]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 0 then begin C := S[P - 0]; VC := VC or ORD(C); end;
    {$ENDIF}

    V.Data[I] := VC;
    Dec (P, CXBytes);
  end;
end;

{
     256-  
  Get a number from the 256-base notation
}
procedure BN_AnsiBase256 (var V: T_BNX; const S: AnsiString; ASign: ShortInt);
var
  L      : IntX;
  I      : LoopInt;
  P      : IntX;
  C      : AnsiChar;
  VC     : UIntX;
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CXBytes);
  BN_Init (V, ASign, VCount);

  P := L;
  for I := 0 to V.HIndex do begin
    VC := 0;

    {$IFDEF CPUX64}
    if P > 7 then begin C := S[P - 7]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 6 then begin C := S[P - 6]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 5 then begin C := S[P - 5]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 4 then begin C := S[P - 4]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 3 then begin C := S[P - 3]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 2 then begin C := S[P - 2]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 1 then begin C := S[P - 1]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 0 then begin C := S[P - 0]; VC := VC or ORD(C); end;
    {$ELSE}
    if P > 3 then begin C := S[P - 3]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 2 then begin C := S[P - 2]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 1 then begin C := S[P - 1]; VC := VC or ORD(C); VC := VC shl 8; end;
    if P > 0 then begin C := S[P - 0]; VC := VC or ORD(C); end;
    {$ENDIF}

    V.Data[I] := VC;
    Dec (P, CXBytes);
  end;
end;

{
     256-  
  Convert the number to 256-base notation
}
function BN_Base256 (const A: T_BNX): string;
var
  L      : IntX;
  I      : LoopInt;
  P      : IntX;
  VC     : UIntX;
  VB     : Byte;
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 3; // VCount * 8;
  {$ELSE}
  L := VCount shl 2; // VCount * 4;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  SetLength (Result, L);
  P := L;
  for I := 0 to VCount - 1 do begin
    VC := A.Data[I];

    {$IFDEF CPUX64}
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); // VC := VC shr 8;
    {$ELSE}
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := CHR(VB); Dec (P); // VC := VC shr 8;
    {$ENDIF}
  end;
end;

{
     256-  
  Convert the number to 256-base notation
}
function BN_AnsiBase256 (const A: T_BNX): AnsiString;
var
  L      : IntX;
  I      : LoopInt;
  P      : IntX;
  VC     : UIntX;
  VB     : Byte;
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 3; // VCount * 8;
  {$ELSE}
  L := VCount shl 2; // VCount * 4;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  SetLength (Result, L);
  P := L;
  for I := 0 to VCount - 1 do begin
    VC := A.Data[I];

    {$IFDEF CPUX64}
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); // VC := VC shr 8;
    {$ELSE}
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); VC := VC shr 8;
    VB := VC and $FF; Result[P] := AnsiChar(VB); Dec (P); // VC := VC shr 8;
    {$ENDIF}
  end;
end;

{
     
  Get a number from the string
}
procedure BN_String (var V: T_BNX; const S: string; ASign: ShortInt);
var
  L      : IntX;
  {$IFDEF UNICODE}
  Sa     : AnsiString;
  {$ENDIF}
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CXBytes);
  BN_Init (V, ASign, VCount);

  {$IFDEF UNICODE}
  Sa := AnsiString(S);
  move (PAnsiChar(Sa)^, V.PLData, L);
  // CopyMemory (V.PLData, PAnsiChar(Sa), L);
  {$ELSE}
  move (PAnsiChar(S)^, V.PLData, L);
  // CopyMemory (V.PLData, PAnsiChar(S), L);
  {$ENDIF}
end;

{
     
  Get a number from the string
}
procedure BN_AnsiString (var V: T_BNX; const S: AnsiString; ASign: ShortInt);
var
  L      : IntX;
  VCount : IntX;
begin
  L := Length (S);
  if L <= 0 then begin
    BN_AssignZero (V);
    Exit;
  end;

  VCount := cdiv (L, CXBytes);
  BN_Init (V, ASign, VCount);

  move (PAnsiChar(S)^, V.PLData, L);
  // CopyMemory (V.PLData, PAnsiChar(S), L);
end;

{
     
  Convert the number to string
}
function BN_String (const A: T_BNX): string;
var
  L      : IntX;
  {$IFDEF UNICODE}
  Sa     : AnsiString;
  {$ENDIF}
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 3; // VCount * 8;
  {$ELSE}
  L := VCount shl 2; // VCount * 4;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  {$IFDEF UNICODE}
  SetLength (Sa, L);
  move (A.PLData, PAnsiChar(Sa)^, L);
  // CopyMemory (PAnsiChar(Sa), A.PLData, L);
  Result := string(Sa);
  {$ELSE}
  SetLength (Result, L);
  move (A.PLData, PAnsiChar(Result)^, L);
  // CopyMemory (PAnsiChar(Result), A.PLData, L);
  {$ENDIF}
end;

{
     
  Convert the number to string
}
function BN_AnsiString (const A: T_BNX): AnsiString;
var
  L      : IntX;
  VCount : IntX;
begin
  VCount := BN_DataCount (A);
  {$IFDEF CPUX64}
  L := VCount shl 3; // VCount * 8;
  {$ELSE}
  L := VCount shl 2; // VCount * 4;
  {$ENDIF}

  if L <= 0 then begin
    Result := '';
    Exit;
  end;

  SetLength (Result, L);
  move (A.PLData, PAnsiChar(Result)^, L);
  // CopyMemory (PAnsiChar(Result), A.PLData, L);
end;

end.

