unit mxBNMathRandom;

// Binuar 1.2.1
// Big integer number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{$I mxBNMath.inc}

interface

uses
  mxBNMathTypes;

type
  P_BBS_Random_Data = ^T_BBS_Random_Data;
  T_BBS_Random_Data = record
    LBits : IntX;
    RBits : UIntX; // random bits
    Rmask : UIntX; // random bits mask
    p     : T_BNX;
    q     : T_BNX;
    m     : T_BNX;
    x     : T_BNX;
  end;

function  rdrand: UIntX;
function  rdseed: UIntX;

function  BN_Randomize (ASeed: UIntX = 0): UIntX;
function  BN_Random: UIntX; overload;
procedure BN_Random (var V: T_BNX); overload;
function  BN_Random (var V: T_BNX; AHmask: UIntX): IntX; overload;
procedure BN_Random (var V: T_BNX; ACount: IntX); overload;

// CPU generator
function  BN_CPU_Random: UIntX; overload;
procedure BN_CPU_Random (var V: T_BNX); overload;

// Linear congruential generator
function  BN_LCG_Randomize (ASeed: UIntX = 0): UIntX;
function  BN_LCG_Random: UIntX; overload;
procedure BN_LCG_Random (var V: T_BNX); overload;

// Blum-Blum-Shub generator
procedure BN_BBS_Random_Init (var AData: T_BBS_Random_Data; ABits: IntX; AForce: Boolean); overload;
procedure BN_BBS_Random_Init (ABits: IntX; AForce: Boolean); overload;
procedure BN_BBS_Random_Clear (var AData: T_BBS_Random_Data); overload;
procedure BN_BBS_Random_Clear; overload;
function  BN_BBS_Random (var AData: T_BBS_Random_Data): UIntX; overload;
function  BN_BBS_Random: UIntX; overload;
procedure BN_BBS_Random (var AData: T_BBS_Random_Data; var V: T_BNX); overload;
procedure BN_BBS_Random (var V: T_BNX); overload;

implementation

uses
  mxBNMath, mxBNMathProcs, mxBNMathUtils;

var
  // CPU Random Instructions Flags
  FCPU_RDRAND : Boolean = false;
  ECPU_RDRAND : Boolean = false;
  FCPU_RDSEED : Boolean = false;
  ECPU_RDSEED : Boolean = false;

var
{$IFDEF CPUX64}
  // MMIX by Donald Knuth
  LCG_a : UIntX = 6364136223846793005;
  LCG_c : UIntX = 1442695040888963407;
  LCG_x : UIntX = 0;
{$ELSE}
  // Numerical Recipes
  LCG_a : UIntX = 1664525;
  LCG_c : UIntX = 1013904223;
  LCG_x : UIntX = 0;
{$ENDIF}

var
  BBS_Data : T_BBS_Random_Data;

(*
     // cpuid
        DB      $0F, $A2

     // rdtsc
        DB      $0F, $31

     // rdrand  rax
        DB      $48, $0F, $C7, $F0

     // rdrand  eax
        DB      $0F, $C7, $F0

     // rdseed  rax
        DB      $48, $0F, $C7, $F8

     // rdseed  eax
        DB      $0F, $C7, $F8
*)

// [+] 1.2.1
{
  RDRAND - Read Random Number
  CF=1 indicates that the data in the destination is valid
}
function rdrand: UIntX;
{$IFDEF CPUX64}
asm
       .NOFRAME
  @RD:
     // rdrand  rax
        DB      $48, $0F, $C7, $F0
        jnc     @RD
end;
{$ELSE}
asm
  @RD:
     // rdrand  eax
        DB      $0F, $C7, $F0
        jnc     @RD
end;
{$ENDIF}

// [+] 1.2.1
{
  RDSEED - Read Random SEED
  CF=1 indicates that the data in the destination is valid
}
function rdseed: UIntX;
{$IFDEF CPUX64}
asm
       .NOFRAME
  @RD:
     // rdseed  rax
        DB      $48, $0F, $C7, $F8
        jnc     @RD
end;
{$ELSE}
asm
  @RD:
     // rdseed  eax
        DB      $0F, $C7, $F8
        jnc     @RD
end;
{$ENDIF}

// [+] 1.2.1
{
}
function BN_Randomize (ASeed: UIntX = 0): UIntX;
begin
  if ASeed = 0 then begin
    if FCPU_RDSEED then ASeed := rdseed;
  end;
  Result := BN_LCG_Randomize (ASeed);
end;

// [+] 1.2.1
{
}
function BN_Random: UIntX;
begin
  if FCPU_RDRAND then begin
    Result := BN_CPU_Random;
    Exit;
  end;

  Result := BN_LCG_Random;
end;

// [+] 1.2.1
{
  V = Random
}
procedure BN_Random (var V: T_BNX);
begin
  if FCPU_RDRAND then begin
    BN_CPU_Random (V);
    Exit;
  end;

  BN_LCG_Random (V);
end;

// [+] 1.2.1
{
  V = Random
}
function BN_Random (var V: T_BNX; AHmask: UIntX): IntX;
begin
  BN_Random (V);

  // apply mask to upper bits
  V.PHData^ := V.PHData^ and AHmask;
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  V = Random
}
procedure BN_Random (var V: T_BNX; ACount: IntX);
begin
  BN_Resize (V, ACount);

  if FCPU_RDRAND then begin
    BN_CPU_Random (V);
    Exit;
  end;

  BN_LCG_Random (V);
end;

// [+] 1.2.1
{
  CPU:RDRAND
  CF=1 indicates that the data in the destination is valid
}
function BN_CPU_Random: UIntX;
{$IFDEF CPUX64}
asm
       .NOFRAME
  @RD:
     // rdrand  rax
        DB      $48, $0F, $C7, $F0
        jnc     @RD
end;
{$ELSE}
asm
  @RD:
     // rdrand  eax
        DB      $0F, $C7, $F0
        jnc     @RD
end;
{$ENDIF}

// [+] 1.2.1
{
  CPU:RDRAND
  CF=1 indicates that the data in the destination is valid

  V = Random
}
procedure BN_CPU_Random (var V: T_BNX);
{$IFDEF CPUX64}
// rcx | @V |
asm
       .NOFRAME

        mov     r11, [rcx].T_BNX.Count
        cmp     r11, 0              // r11 = V.Count
        jle     @R

  @L:
        mov     rdx, [rcx].T_BNX.PLData
  @A:                               // for I := 0 to HIndex ...
  @RD:
     // rdrand  rax
        DB      $48, $0F, $C7, $F0
        jnc     @RD
        mov     [rdx], rax          // V.Data[I] := random
  @D:
        add     rdx, CXBytes        // I := I + 1
        dec     r11
        jnz     @A

  @R:
end;
{$ELSE}
// eax | @V |
asm
        mov     ecx, [eax].T_BNX.Count
        cmp     ecx, 0              // ecx = V.Count
        jle     @R

  @L:
        mov     edx, [eax].T_BNX.PLData
  @A:                               // for I := 0 to HIndex ...
  @RD:
     // rdrand  eax
        DB      $0F, $C7, $F0
        jnc     @RD
        mov     [edx], eax          // V.Data[I] := random
  @D:
        add     edx, CXBytes        // I := I + 1
        dec     ecx
        jnz     @A

  @R:
end;
{$ENDIF}

// [+] 1.2.1
{
}
function BN_LCG_Randomize (ASeed: UIntX = 0): UIntX;
{$IFDEF CPUX64}
// rcx | ASeed |
asm
       .NOFRAME

        mov     rax, rcx
        test    rax, rax
        jnz     @C

        rdtsc                       // eax:edx = processor's time-stamp counter
        xor     rdx, rsp
        shl     rdx, 32
        or      rax, rdx
  @C:
        lea     rdx, LCG_x          // rdx = @LCG_x
        mov     [rdx], rax          // LCG_x := seed

  @R:
end;
{$ELSE}
// eax | ASeed |
asm
        test    eax, eax
        jnz     @C

        {$IFDEF __RDTSC}
        rdtsc                       // eax:edx = processor's time-stamp counter
        {$ELSE}
     // rdtsc
        DB      $0F, $31
        {$ENDIF}
        xor     eax, edx
  @C:
        lea     edx, LCG_x          // edx = @LCG_x
        mov     [edx], eax          // LCG_x := seed

  @R:
end;
{$ENDIF}

// [+] 1.2.1
{
    
  Linear congruential generator

  x(n+1) = (a*x(n)+c) mod m
  m = 2^32|64
}
function BN_LCG_Random: UIntX;
{$IFDEF CPUX64}
asm
       .NOFRAME

        mov     rax, LCG_x          // rax = x(n)
        mul     LCG_a               // mul rax, LCG_a
        add     rax, LCG_c          // rax = x(n+1)
        lea     rdx, LCG_x          // rdx = @LCG_x
        mov     [rdx], rax          // LCG_x := x(n+1)
end;
{$ELSE}
asm
        mov     eax, LCG_x          // eax = x(n)
        mul     LCG_a               // mul eax, LCG_a
        add     eax, LCG_c          // eax = x(n+1)
        lea     edx, LCG_x          // edx = @LCG_x
        mov     [edx], eax          // LCG_x := x(n+1)
end;
{$ENDIF}

// [+] 1.2.1
{
    
  Linear congruential generator

  x(n+1) = (a*x(n)+c) mod m
  m = 2^32|64
}
procedure BN_LCG_Random (var V: T_BNX);
{$IFDEF CPUX64}
// rcx | @V |
asm
       .NOFRAME

        mov     r11, [rcx].T_BNX.Count
        cmp     r11, 0              // r11 = V.Count
        jle     @R

  @L:
        mov     r10, [rcx].T_BNX.PLData
        mov     r8, LCG_a
        mov     r9, LCG_c
        mov     rax, LCG_x          // rax = x(n)
        lea     rcx, LCG_x          // rcx = @LCG_x
  @A:                               // for I := 0 to HIndex ...
        mul     r8                  // mul rax, r8
        add     rax, r9             // rax = x(n+1)
        mov     [rcx], rax          // LCG_x := x(n+1)
        mov     [r10], rax          // V.Data[I] := x(n+1)
  @D:
        add     r10, CXBytes        // I := I + 1
        dec     r11
        jnz     @A

  @R:
end;
{$ELSE}
// eax | @V |
asm
        push    ebx
        push    esi
        push    edi

        mov     ecx, [eax].T_BNX.Count
        cmp     ecx, 0              // ecx = V.Count
        jle     @R

  @L:
        mov     ebx, [eax].T_BNX.PLData
        mov     esi, LCG_a
        mov     edi, LCG_c
        mov     eax, LCG_x          // eax = x(n)
  @A:                               // for I := 0 to HIndex ...
        mul     esi                 // mul eax, esi
        add     eax, edi            // eax = x(n+1)
        lea     edx, LCG_x          // edx = @LCG_x
        mov     [edx], eax          // LCG_x := x(n+1)
        mov     [ebx], eax          // V.Data[I] := x(n+1)
  @D:
        add     ebx, CXBytes        // I := I + 1
        dec     ecx
        jnz     @A

  @R:
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
  RDRAND : CPU_Rep.f_1_ECX_[30]
  RDSEED : CPU_Rep.f_7_EBX_[18]
}
procedure GetCPURandomFlags;
{$IFDEF CPUX64}
asm
       .NOFRAME

        push    rbx

        xor     rbx, rbx
        lea     r8, FCPU_RDRAND     // r8 = @FCPU_RDRAND
        mov     [r8], bl            // FCPU_RDRAND := FALSE
        lea     r9, FCPU_RDSEED     // rdx = @FCPU_RDSEED
        mov     [r9], bl            // FCPU_RDSEED := FALSE

  @RDRAND:
        mov     eax, 1
        cpuid
        test    ecx, $40000000      // ECX:30 - RDRAND bit
        jz      @RDRAND_OUT
        mov     bl, 1
        mov     [r8], bl            // FCPU_RDRAND := TRUE
  @RDRAND_OUT:

  @RDSEED:
        mov     eax, 7
        cpuid
        test    ebx, $00040000      // EBX:18 - RDSEED bit
        jz      @RDSEED_OUT
        mov     bl, 1
        mov     [r9], bl            // FCPU_RDSEED := TRUE
  @RDSEED_OUT:

  @R:
        pop     rbx
end;
{$ELSE}
asm
        push    ebx
        push    esi
        push    edi

        xor     ebx, ebx
        lea     edi, FCPU_RDRAND    // edi = @FCPU_RDRAND
        mov     [edi], bl           // FCPU_RDRAND := FALSE
        lea     esi, FCPU_RDSEED    // esi = @FCPU_RDSEED
        mov     [esi], bl           // FCPU_RDSEED := FALSE

  @RDRAND:
        mov     eax, 1
        {$IFDEF __CPUID}
        cpuid
        {$ELSE}
     // cpuid
        DB      $0F, $A2
        {$ENDIF}
        test    ecx, $40000000      // ECX:30 - RDRAND bit
        jz      @RDRAND_OUT
        mov     bl, 1
        mov     [edi], bl           // FCPU_RDRAND := TRUE
  @RDRAND_OUT:

  @RDSEED:
        mov     eax, 7
        {$IFDEF __CPUID}
        cpuid
        {$ELSE}
     // cpuid
        DB      $0F, $A2
        {$ENDIF}
        test    ebx, $00040000      // EBX:18 - RDSEED bit
        jz      @RDSEED_OUT
        mov     bl, 1
        mov     [esi], bl           // FCPU_RDSEED := TRUE
  @RDSEED_OUT:

  @R:
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
}
procedure BN_BBS_Random_Init (var AData: T_BBS_Random_Data; ABits: IntX;
  AForce: Boolean);
var
  VBits  : UIntX;
  V_shr  : IntX;
  V_mask : UIntX;
begin
  if not AForce then begin
    if AData.RBits <> 0 then Exit;
  end;

  VBits := CXBits;
  if ABits > 0 then begin
    VBits := CXBits * (ABits div CXBits);
    if VBits = 0 then VBits := CXBits;
  end;

  AData.LBits := VBits;

  BN_GenPrime_3m4 (AData.p, VBits, 64);
  BN_GenPrime_3m4 (AData.q, VBits, 64);
  // p = 3 (mod 4)
  // q = 3 (mod 4)

  BN_mul (AData.m, AData.p, AData.q);
  // m = p*q

  // m = (2^n)*(1+eps), eps<1
  // Log2(m) = Log2(2^n) + Log2(1+eps) = n
  // Log2(Log2(m)) = Log2(n)
  VBits := bsr (BN_bsr (AData.m));
  AData.RBits := VBits;
  AData.Rmask := CXmask shr (CXBits - VBits);

  // generate random x < m
  BN_Random (AData.x, AData.m.Count);
  if AData.x.Count = AData.m.Count then begin
    V_shr := CXBits - bsr (AData.m.PHData^);
    V_mask := 0;
    if V_shr < CXBits then V_mask := CXmask shr V_shr;

    // clear upper bits to make x < m
    AData.x.PHData^ := AData.x.PHData^ and V_mask;
    BN_Pack (AData.x);
  end;

  // get x co-prime to m
  BN_GetLCoPrime (AData.x, AData.m);
end;

// [+] 1.2.1
{
}
procedure BN_BBS_Random_Init (ABits: IntX; AForce: Boolean);
begin
  BN_BBS_Random_Init (BBS_Data, ABits, AForce);
end;

// [+] 1.2.1
{
}
procedure BN_BBS_Random_Clear (var AData: T_BBS_Random_Data);
begin
  AData.LBits := 0;
  AData.RBits := 0;
  AData.Rmask := 0;

  BN_Free (AData.p);
  BN_Free (AData.q);
  BN_Free (AData.m);
  BN_Free (AData.x);
end;

// [+] 1.2.1
{
}
procedure BN_BBS_Random_Clear;
begin
  BN_BBS_Random_Clear (BBS_Data);
end;

// [+] 1.2.1
{
  Blum-Blum-Shub generator
  x(n+1) = x(n)^2 mod m
}
function BN_BBS_Random (var AData: T_BBS_Random_Data): UIntX;
var
  I : LoopInt;
  C : IntX;
  V : T_BNX;
  R : UIntX;
  D : UIntX;
begin
  Result := 0;
  if AData.Rmask = 0 then Exit;

  C := cdiv (CXBits, AData.RBits);
  for I := 0 to C - 1 do begin
    BN_sqr (V, AData.x);
    BN_mod (AData.x, V, AData.m);
    // x = x^2 mod m

    R := AData.x.Data[0] and AData.Rmask; // R = random bits
    D := Result shl AData.RBits;
    Result := D or R;
  end;

  BN_LFree (V);
end;

// [+] 1.2.1
{
  Blum-Blum-Shub generator
  x(n+1) = x(n)^2 mod m
}
function BN_BBS_Random: UIntX;
begin
  Result := BN_BBS_Random (BBS_Data);
end;

// [+] 1.2.1
{
  Blum-Blum-Shub generator
  x(n+1) = x(n)^2 mod m
}
procedure BN_BBS_Random (var AData: T_BBS_Random_Data; var V: T_BNX);
var
  I : LoopInt;
begin
  for I := 0 to V.Count - 1 do begin
    V.Data[I] := BN_BBS_Random (AData);
  end;
end;

// [+] 1.2.1
{
  Blum-Blum-Shub generator
  x(n+1) = x(n)^2 mod m
}
procedure BN_BBS_Random (var V: T_BNX);
begin
  BN_BBS_Random (BBS_Data, V);
end;

// [+] 1.2.1
{
}
procedure Enter;
var
  VSeed : UIntX;
begin
  GetCPURandomFlags;

  if FCPU_RDRAND then begin
    try
      rdrand;
    except
      FCPU_RDRAND := false;
      ECPU_RDRAND := true;
    end;
  end;

  VSeed := 0;
  if FCPU_RDSEED then begin
    try
      VSeed := rdseed;
    except
      VSeed := 0;
      FCPU_RDSEED := false;
      ECPU_RDSEED := true;
    end;
  end;

  BN_LCG_Randomize (VSeed);
  BN_BBS_Random_Clear;
end;

initialization
  mxBNMathRandom.Enter;

end.

