unit mxBNMathTypes;

// Binuar 1.2.1
// Big integer number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{$I mxBNMath.inc}

interface

const
  {$IFDEF CPUX64}
  CXBits  = 64;
  CXBytes = 8;
  CXBiSh  = 6; // CXBits  = 2^CXBiSh
  CXBySh  = 3; // CXBytes = 2^CXBySh
  CXmax   = $FFFFFFFFFFFFFFFF;
  CXmask  = $FFFFFFFFFFFFFFFF;
  CXHiBit = $8000000000000000;
  {$ELSE}
  CXBits  = 32;
  CXBytes = 4;
  CXBiSh  = 5; // CXBits  = 2^CXBiSh
  CXBySh  = 2; // CXBytes = 2^CXBySh
  CXmax   = $FFFFFFFF;
  CXmask  = $FFFFFFFF;
  CXHiBit = $80000000;
  {$ENDIF}

const
  {$IFDEF CPUX64}
  Karatsuba_mul_TCount = 80;
  Karatsuba_sqr_TCount = 60;
  BZ_DIV_LIMIT         = 60; // Burnikel-Ziegler DIV_LIMIT
  {$ELSE}
  Karatsuba_mul_TCount = 80;
  Karatsuba_sqr_TCount = 60;
  BZ_DIV_LIMIT         = 60; // Burnikel-Ziegler DIV_LIMIT
  {$ENDIF}

type Int8   = ShortInt;
type UInt8  = Byte;
type Int16  = SmallInt;
type UInt16 = Word;
type Int32  = Integer;
type UInt32 = Cardinal;

{$IFNDEF __PBoolean}
type PBoolean = ^Boolean;
{$ENDIF}

type
  {$IFDEF CPUX64}
  PIntX = ^IntX;
  IntX = Int64;

  PUIntX = ^UIntX;
  UIntX = UInt64;

  {$IFDEF LOOP_INTX}
  LoopInt = IntX;
  LoopUInt = UIntX;
  {$ELSE}
  LoopInt = Int32;
  LoopUInt = UInt32;
  {$ENDIF}
  {$ELSE}
  PIntX = ^IntX;
  IntX = Int32;

  PUIntX = ^UIntX;
  UIntX = UInt32;

  LoopInt = IntX;
  LoopUInt = UIntX;
  {$ENDIF}

const
  CXDataCount = maxInt div SizeOf(UintX);

type
  PUIntXData = ^TUIntXData;
  TUIntXData = array of UIntX;
  (*
  PUIntXData = ^TUIntXData;
  TUIntXData = array [0..CXDataCount-1] of UIntX;
  // *)

type
  P_BNX = ^T_BNX;
  T_BNX = record
    Sign   : ShortInt; // -1 0 +1
    Count  : IntX;     //    Data
                       // Number of Data array elements
    HIndex : IntX;     // HIGH(Data)
    PLData : PUIntX;   // @Data[0]
    PHData : PUIntX;   // @Data[HIndex]
    Data   : TUIntXData;
  end;

type
  T_BNX_Array = array of T_BNX;

const
  // Unique bytes of a^2, 0 <= a < 255
  // $_0, $_1, $_4, $_9
  CSqrByte : array [0..43] of Byte = (
    $00, $01, $04, $09, $10, $11, $19, $21,
    $24, $29, $31, $39, $40, $41, $44, $49,
    $51, $59, $61, $64, $69, $71, $79, $81,
    $84, $89, $90, $91, $99, $A1, $A4, $A9,
    $B1, $B9, $C1, $C4, $C9, $D1, $D9, $E1,
    $E4, $E9, $F1, $F9
  );

  // Unique bytes of a^2, 0 <= a < 255, a is odd
  // $_1, $_9
  CSqrOddByte : array [0..31] of Byte = (
    $01, $09, $11, $19, $21, $29, $31, $39,
    $41, $49, $51, $59, $61, $69, $71, $79,
    $81, $89, $91, $99, $A1, $A9, $B1, $B9,
    $C1, $C9, $D1, $D9, $E1, $E9, $F1, $F9
  );

  // Unique bytes of a^2, 0 <= a < 255, a is even
  // $_0, $_4
  // $00, $10, $40, $90
  // $04, $24, $44, $64, $84, $A4, $C4, $E4
  CSqrEvenByte : array [0..11] of Byte = (
    $00, $04, $10, $24, $40, $44, $64, $84,
    $90, $A4, $C4, $E4
  );

const
  // Unique bytes of a^4, 0 <= a < 255
  // $_0, $_1
  CP4Byte : array [0..17] of Byte = (
    $00, $01, $10, $11, $21, $31, $41, $51,
    $61, $71, $81, $91, $A1, $B1, $C1, $D1,
    $E1, $F1
  );

  // Unique bytes of a^4, 0 <= a < 255, a is odd
  // $_1
  CP4OddByte : array [0..15] of Byte = (
    $01, $11, $21, $31, $41, $51, $61, $71,
    $81, $91, $A1, $B1, $C1, $D1, $E1, $F1
  );

  // Unique bytes of a^4, 0 <= a < 255, a is even
  // $_0
  CP4EvenByte : array [0..1] of Byte = (
    $00, $10
  );

implementation

end.

